/*
    Example: VT1300 GPIO Control
    Description: An example device that allows control of digital pin 3.  The device type is a router.
*/

/* ------------------------------------------------------------------
   Includes
   ------------------------------------------------------------------
*/
#include <VT1300_Shield.h>

/* ------------------------------------------------------------------
   Class Instances
   ------------------------------------------------------------------
*/
CC2530 mycc2530;                                                      // Library VT1300 Shield Class Instance

/* ------------------------------------------------------------------
   Pins
   ------------------------------------------------------------------
*/

#define PIN_CTS     9                                                 // (P0.5) RT
#define PIN_RTS     8                                                 // (P0.4) CT
#define PIN_RES     7                                                 // (RST) Reset

#define PIN_D3      3                                                 // Digital Pin 3
#define Button      2                                                 // Commission button

/* ------------------------------------------------------------------
   Global Veriables
   ------------------------------------------------------------------
*/
uint8_t seqCount = 0;                                                 // Used for keeping track of message transaction sequences
uint8_t seqNumber = 0;                                                // Used to match read attribute requests to read attribute responses
uint8_t PIN_D3_Status = 0x00;

/* ------------------------------------------------------------------
   ZCL: Read Attribute Response Commands
   ------------------------------------------------------------------
*/
const uint8_t ModelIdentifier[] = {'V', 'T', '1', '3', '0', '0', 'G', 'P', 'I', 'O', 'C', 'o', 'n', 't', 'r', 'o', 'l'}; // Model Identifier: VT1300GPIOControl
const uint8_t ManufacturerName[] = {'V', 'e', 'r', 't', 'o', 'r', 'i', 'x'};                                             // Manufacturer Name: Vertorix
const uint8_t PowerSource = 4;                                                                                           // Power Source: 0 = Unknown, 1 = mains(single phase), 2 = mains(3 phase), 3 = battery, 4 = DC source.
const uint8_t ApplicationVersion = 1;                                                                                    // Application Version
const uint8_t ZCLVersion = 1;                                                                                            // ZCL version
const uint8_t StackVersion = 2;                                                                                          // Stack version
const uint8_t HWVersion = 1;                                                                                             // Hardware version
const uint8_t Datecode[] = {'2', '0', '2', '1', '0', '2', '0', '3'};                                                     // Date Code
const uint8_t SoftwareBuildID[] = {'1', '2', '0', '0', '-', '0', '0', '0', '1'};                                         // Software Build ID

/********************************************************************
   Setup
 ********************************************************************
*/
void setup()
{
  Serial.begin(115200);
  Serial1.begin(115200);
  delay(100);

  pinMode(PIN_D3, OUTPUT);
  digitalWrite(PIN_D3, LOW);

  // Parameters for CC2530
  mycc2530.SetPANID(0xffff);                                          // PAN ID.  Two bytes set between 0x0000 and 0x3FFF.  Examples: 0x00A1, 0x00A2 or 0x00A3.
  mycc2530.SetLOGICAL_TYPE(0x01);                                     // Device type.  Examples: Coordinator = 0x00, Router = 0x01 or End Device = 0x02
  mycc2530.SetCHANLIST(11);                                           // Wireless Channel. Examples: 11 to 26 or 0xFF All Channels

  Init_CC2530();
}

/********************************************************************
   Loop
 ********************************************************************
*/
void loop()
{
  mycc2530.RECEIVE_SERIAL();                                          // Need to constantly Receive serial data from the CC2530 to the application processor

  if (mycc2530.NEW_DATA())
  {
    if (mycc2530.AF_INCOMING_MSG())                                   // Function returns true if a AF_INCOMING_MSG is received
    { // Process Incoming Messages in this block of code
      mycc2530.LINK_QUALITY();                                        // Function prints the Short Address and Link Quality from a received message.  The link quality is from the last Hop to the receiving device.

      uint8_t ClusterID[2];
      ClusterID[0] = mycc2530.ReceivedBytes[5];
      ClusterID[1] = mycc2530.ReceivedBytes[6];

      if (ClusterID[0] == 0x00 && ClusterID[1] == 0x00)
      {
        ZCL_Interview();                                              // Send ZCL Read attributes response.  This is used in the interview process.
      }
      else if (ClusterID[0] == 0x06 && ClusterID[1] == 0x00)
      {
        ProcessOnOffCluster();                                        // Send ZCL report attributes and default response
      }
    }
  }
}


/********************************************************************
   Other Functions
 ********************************************************************
*/

/* ------------------------------------------------------------------
   Initialise CC2530
   ------------------------------------------------------------------
*/
void Init_CC2530()
{
  mycc2530.POWER_UP();                                                // Power up the CC2530 (wake on reset).
  SetTXPower();

  pinMode(Button, INPUT_PULLUP);
  if (digitalRead(Button) == LOW)
  {
    mycc2530.COMMISSION();                                            // Clears the configuration and network state then writes the new configuration parameters to the CC2530 non-volitile (NV) memory.  This should only be run once on initial setup.
    AF_REGISTER(0x01);                                                // Register Endpoint 1 ZCL
    mycc2530.ZDO_STARTUP_FROM_APP(60000);                             // Starts the CC2530 in the network
    delay(5000);
    AF_REGISTER(0x01);                                                // Register Endpoint 1 ZCL
    mycc2530.ZDO_STARTUP_FROM_APP(60000);                             // Starts the CC2530 in the network
  }

  if (digitalRead(Button) == HIGH)
  {
    AF_REGISTER(0x01);
    mycc2530.ZDO_STARTUP_FROM_APP(60000);                             // Starts the CC2530 in the network
    delay(5000);
  }
}

/* ------------------------------------------------------------------
   Set TX Power Function
   ------------------------------------------------------------------
*/
void SetTXPower()
{
  uint8_t TXPower[] = {0x01, 0x21, 0x14, 0x04};

  mycc2530.WRITE_DATA(TXPower);
  Serial.println("~~TX power set~~");
}

/* ------------------------------------------------------------------
   AF_Register Function
   ------------------------------------------------------------------
*/
void AF_REGISTER(uint8_t EndPoint)
{
  uint8_t Len = 0x0F; // total length after Cmd1 excl. XOR
  uint8_t Cmd0 = 0x24;
  uint8_t Cmd1 = 0x00;
  uint8_t AppEndPoint = EndPoint; // Register End Point
  uint8_t AppProfileID0 = 0x04;
  uint8_t AppProfileID1 = 0x01;
  uint8_t DeviceID0 = 0x00;
  uint8_t DeviceID1 = 0x00;
  uint8_t DeviceVer = 0x01;
  uint8_t LatencyReq = 0x00;
  uint8_t AppNumInClusters = 0x02;
  uint8_t InCluster0 = 0x00;
  uint8_t InCluster1 = 0x00;
  uint8_t InCluster2 = 0x06;
  uint8_t InCluster3 = 0x00;
  uint8_t AppNumOutClusters = 0x01;
  uint8_t OutCluster0 = 0x06;
  uint8_t OutCluster1 = 0x00;
  uint8_t ArraySize = Len + 3;
  // Make array
  uint8_t Data[] = {Len, Cmd0, Cmd1, AppEndPoint, AppProfileID0, AppProfileID1, DeviceID0, DeviceID1, DeviceVer, LatencyReq, AppNumInClusters, InCluster0, InCluster1, InCluster2, InCluster3, AppNumOutClusters, OutCluster0, OutCluster1};
  uint8_t FCS = mycc2530.CALC_FCS(Data, ArraySize);
  Serial.println(F("AF_REGISTER SREQ"));
  // SREQ
  digitalWrite(PIN_RTS, HIGH);
  while (digitalRead(PIN_CTS) == HIGH) {};
  Serial1.write(0xFE);
  for (int i = 0; i < ArraySize; i++)
  {
    Serial1.write(Data[i]);
  }
  Serial1.write(FCS);
  Serial1.flush();
  digitalWrite(PIN_RTS, LOW);
  mycc2530.RECV_CALLBACK(500);
}

/* ------------------------------------------------------------------
   ZCL Read Attributes Response
   ------------------------------------------------------------------
*/
void ZCLFrame_ReadAttributesResponse(uint8_t Data[], uint8_t arraysize)
{
  // ZCL Header
  uint8_t FrameControl = 0x18;                                      // Bitfield that defines the command type and other relevant information in the ZCL command
  uint8_t SequenceNumber = seqNumber;                               // A sequence number used to correlate a ZCL command with a ZCL response.
  uint8_t CommandID = 0x01;                                         // Read Attributes Response
  const uint8_t ZCLHeader[] = {FrameControl, SequenceNumber, CommandID};  // ZCL Header

  // ZCL Payload
  uint8_t ZCLPayload[30];
  memcpy(ZCLPayload, Data, arraysize);

  // ZCL Frame
  uint8_t ZCLFrame[sizeof(ZCLHeader) + sizeof(ZCLPayload)];
  memcpy(ZCLFrame, ZCLHeader, sizeof(ZCLHeader));
  memcpy(ZCLFrame + sizeof(ZCLHeader), ZCLPayload, sizeof(ZCLPayload));
  uint8_t ZCLFrameLength = sizeof(ZCLHeader) + arraysize;

  // Send ZCL Frame
  mycc2530.SetAF_DATA_REQUEST(0x01, 0x01, 0x00, 0x00, 0x00, 0x10, 0x30);
  mycc2530.AF_DATA_REQUEST(0x00, 0x00, ZCLFrame, ZCLFrameLength);
}

/* ------------------------------------------------------------------
   ZCL Interview
   ------------------------------------------------------------------
*/
void ZCL_Interview()
{
  uint8_t ControlField = mycc2530.ReceivedBytes[20];
  seqNumber = mycc2530.ReceivedBytes[21];
  uint8_t Command = mycc2530.ReceivedBytes[22];
  uint8_t Attribute[2];
  Attribute[0] = mycc2530.ReceivedBytes[23];
  Attribute[1] = mycc2530.ReceivedBytes[24];

  if (Attribute[0] == 0x05 && Attribute[1] == 0x00)
  {
    uint8_t Status = 0x00;
    uint8_t DataType = 0x42; // Char String
    uint8_t Length = sizeof(ModelIdentifier);

    uint8_t Payload[5 + Length] = {Attribute[0], Attribute[1], Status, DataType, Length};
    memcpy(Payload + 5, ModelIdentifier, sizeof(ModelIdentifier));

    ZCLFrame_ReadAttributesResponse(Payload, sizeof(Payload));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x04 && Attribute[1] == 0x00)
  {
    uint8_t Status = 0x00;
    uint8_t DataType = 0x42; // Char String
    uint8_t Length = sizeof(ManufacturerName);

    uint8_t Payload[5 + Length] = {Attribute[0], Attribute[1], Status, DataType, Length};
    memcpy(Payload + 5, ManufacturerName, sizeof(ManufacturerName));

    ZCLFrame_ReadAttributesResponse(Payload, sizeof(Payload));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x07 && Attribute[1] == 0x00)
  {
    uint8_t _PowerSource[] = {0x07, 0x00, 0x00, 0x30, PowerSource};
    ZCLFrame_ReadAttributesResponse(_PowerSource, sizeof(_PowerSource));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x01 && Attribute[1] == 0x00)
  {
    uint8_t _ApplicationVersion[] = {0x01, 0x00, 0x00, 0x20, ApplicationVersion};
    ZCLFrame_ReadAttributesResponse(_ApplicationVersion, sizeof(_ApplicationVersion));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x00 && Attribute[1] == 0x00)
  {
    uint8_t _ZCLVersion[] = {0x00, 0x00, 0x00, 0x20, ZCLVersion};
    ZCLFrame_ReadAttributesResponse(_ZCLVersion, sizeof(_ZCLVersion));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x02 && Attribute[1] == 0x00)
  {
    uint8_t _StackVersion[] = {0x02, 0x00, 0x00, 0x20, StackVersion};
    ZCLFrame_ReadAttributesResponse(_StackVersion, sizeof(_StackVersion));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x03 && Attribute[1] == 0x00)
  {
    uint8_t _HWVersion[] = {0x03, 0x00, 0x00, 0x20, HWVersion};
    ZCLFrame_ReadAttributesResponse(_HWVersion, sizeof(_HWVersion));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x06 && Attribute[1] == 0x00)
  {
    uint8_t Status = 0x00;
    uint8_t DataType = 0x42; // Char String
    uint8_t Length = sizeof(Datecode);

    uint8_t Payload[5 + Length] = {Attribute[0], Attribute[1], Status, DataType, Length};
    memcpy(Payload + 5, Datecode, sizeof(Datecode));

    ZCLFrame_ReadAttributesResponse(Payload, sizeof(Payload));
    mycc2530.RECV_CALLBACK(500);
  }
  else if (Attribute[0] == 0x00 && Attribute[1] == 0x40)
  {
    uint8_t Status = 0x00;
    uint8_t DataType = 0x42; // Char String
    uint8_t Length = sizeof(SoftwareBuildID);

    uint8_t Payload[5 + Length] = {Attribute[0], Attribute[1], Status, DataType, Length};
    memcpy(Payload + 5, SoftwareBuildID, sizeof(SoftwareBuildID));

    ZCLFrame_ReadAttributesResponse(Payload, sizeof(Payload));
    mycc2530.RECV_CALLBACK(500);
  }
}

/* ------------------------------------------------------------------
   ZCL Report GPIO Status
   ------------------------------------------------------------------
*/

void ZCLFrame_ReportAttributes_PIN_D3_Status()
{
  // ZCL Header
  uint8_t FrameControl = 0x18;                                      // Bitfield that defines the command type and other relevant information in the ZCL command
  uint8_t SequenceNumber = seqCount;                                // A sequence number used to correlate a ZCL command with a ZCL response.
  seqCount++;
  uint8_t CommandID = 0x0a;                                         // Report Attributes
  uint8_t ZCLHeader[] = {FrameControl, SequenceNumber, CommandID};  // ZCL Header

  // ZCL Payload
  uint8_t AttributeID[2] = {0x00, 0x00};                            // Attribute ID: OnOff
  uint8_t DataType = 0x10;                                          // Data Type: Boolean
  uint8_t OnOff = PIN_D3_Status;                                    // On/Off Status of Digital Pin D3. 0x00 = off, 0x01 = on
  uint8_t ZCLPayload[] = {AttributeID[0], AttributeID[1], DataType, OnOff}; // ZCL Payload

  // ZCL Frame
  uint8_t ZCLFrame[sizeof(ZCLHeader) + sizeof(ZCLPayload)];
  memcpy(ZCLFrame, ZCLHeader, sizeof(ZCLHeader));
  memcpy(ZCLFrame + sizeof(ZCLHeader), ZCLPayload, sizeof(ZCLPayload));

  // Send ZCL Frame
  mycc2530.SetAF_DATA_REQUEST(0x01, 0x01, 0x06, 0x00, 0x00, 0x10, 0x30);  // Destination EP, Source EP, ClusterID, ClusterID, Trans ID, Options, Radius
  mycc2530.AF_DATA_REQUEST(0x00, 0x00, ZCLFrame, sizeof(ZCLFrame));       // NwkAddr, NwkAddr, Payload to Send, Length of Payload
}

/* ------------------------------------------------------------------
   ZCL Default Response
   ------------------------------------------------------------------
*/

void ZCLFrame_ReportAttributes_DefaultResponse()
{
  // ZCL Header
  uint8_t FrameControl = 0x18;                                      // Bitfield that defines the command type and other relevant information in the ZCL command
  uint8_t SequenceNumber = seqNumber;                               // A sequence number used to correlate a ZCL command with a ZCL response.
  uint8_t CommandID = 0x0b;                                         // CommandID: Default Response
  uint8_t ZCLHeader[] = {FrameControl, SequenceNumber, CommandID};  // ZCL Header

  // ZCL Payload
  uint8_t OnOff = PIN_D3_Status;                                    // CommandID: On/Off Status of Digital Pin D3
  uint8_t Status = 0x00;                                            // Status: 0=success
  uint8_t ZCLPayload[] = {OnOff, Status};                           // ZCL Payload

  // ZCL Frame
  uint8_t ZCLFrame[sizeof(ZCLHeader) + sizeof(ZCLPayload)];
  memcpy(ZCLFrame, ZCLHeader, sizeof(ZCLHeader));
  memcpy(ZCLFrame + sizeof(ZCLHeader), ZCLPayload, sizeof(ZCLPayload));

  // Send ZCL Frame
  mycc2530.SetAF_DATA_REQUEST(0x01, 0x01, 0x06, 0x00, 0x00, 0x10, 0x30);  // Destination EP, Source EP, ClusterID, ClusterID, Trans ID, Options, Radius
  mycc2530.AF_DATA_REQUEST(0x00, 0x00, ZCLFrame, sizeof(ZCLFrame));       // NwkAddr, NwkAddr, Payload to Send, Length of Payload
}

/* ------------------------------------------------------------------
   ZCL Process OnOff Cluster
   ------------------------------------------------------------------
*/
void ProcessOnOffCluster()
{
  uint8_t ControlField = mycc2530.ReceivedBytes[20];
  seqNumber = mycc2530.ReceivedBytes[21];
  uint8_t Command = mycc2530.ReceivedBytes[22];                   // On Off Value

  digitalWrite(PIN_D3, Command);                                  // Set PIN_D3 to On or Off
  PIN_D3_Status = digitalRead(PIN_D3);
  ZCLFrame_ReportAttributes_PIN_D3_Status();                      // Report the Status of GPIO Pin Digital 3
  mycc2530.RECV_CALLBACK(500);
  ZCLFrame_ReportAttributes_DefaultResponse();                    // Send mandatory Default Response
  mycc2530.RECV_CALLBACK(500);
}
